open Util
open Std
open CamomileLibrary.Default.Camomile
open Printf
module Cm = CaseMap.Make (UTF16)
module Ul = ULine.Make (UTF16)
module Ce = CharEncoding
module Re = UReStr.Make (UTF16)

let start_time = iof (Unix.time ());;

let rng = Gsl_rng.make (Gsl_rng.default ());;
Random.self_init ();;
let seed = Random.nativeint (Nativeint.of_int max_int);;
Gsl_rng.set rng seed;;
let set_seed n =
  Gsl_rng.set rng (Nativeint.of_int n);
  Random.init n
;;
(*set_seed 0;;*)
let print_utf16 u = 
   let output = new Ce.out_channel Ce.utf16le stdout in
   UTF16.iter output#put u;
   output#flush ()
;;  
let utf16_of_str s = UTF16.init (S.length s) (fun i -> UChar.of_char s.[i]);;
let utf16_to_str u = S.init (UTF16.length u) (fun i -> UChar.char_of (UTF16.get u i));;

let pi = 4.0 *. atan 1.0;;
let log_fact n = 
   n *. log n -. n +. log ((n *. (1. +. 4. *. n *. (1. +. 2.*.n))) /. 6.) +. log pi /. 2.
;;

type omega     = float array
type mapping   = { words : UTF16.t array; pos : string array}
type word      = {
   id                : int;
   g_pos             : int;
   mutable p_pos     : int;
   mutable x_pos     : int;
   mutable probs     : float array;
   zindex            : int array}
;;                   
type data      = {
   langs             : string array;
   sents             : word array array array;         (* sentences.(i).(j) = language i, sentence j, array of words *)
   maps              : mapping array;                  (* mapping.(i) = language i *)
   lexs              : int array array array;          (* lexicon.(l).(w) = language l, word w : array of POS's *) 
   n                 : float;                          (* n number of latent variables (cliques) for languages *) 
   testpoints        : int array;
   endpoints         : int array;
   suppoints         : (int * int) option array;
   z_values          : int array;                      (* z_values.(i) is z_value for ith latent z, one unique index for each clique*)
   z_locations       : (int * int) array array;        (* z_locations.(i).(l) = (s,w) location of the leg of ith z in language l. negative values when it is not there *)
   }
;;   
type model     = {
   theta             : float array;                    (* theta.(i) = hyper param for emissions of language i *) 
   eta               : float array;                    (* eta.(l) = hyperparameters for transition dirichlet prior *)
   zeta              : float array;                    (* zeta.(l) = hyperparameters for omega base distribution dirichlet *)
   
   emcount           : float array array array;        (* emcount.(l).(t) = array of word counts for language l, tag t *)  
   monocount         : float array array;              (* monocount.(l) = array of tag counts *) 
   trcount           : float array array array array;  (* trcount.(l).(t1).(t2) = array of subsequent tag counts *)  
   bicount           : float array array array;        (* bicount.(l).(t) = array of subsequent tag counts *) 

   emcountx          : float array array array;        (* similar counts, but for x_pos instead of p_pos *)
   monocountx        : float array array;
   trcountx          : float array array array array; 
   bicountx          : float array array array;

   tagwordcount      : float array array;              (* tagwordcount.(l).(t) = number of words that can with tag t*)

   mutable alpha     : float ;                         (* concentration parameter for omegas *)
   zcount            : float array;                    (* zcount.(z) = count of latent value z*)
   mutable ztotal    : float ;
   activez           : (int Hs.t);                     (* activez = set of z values currently in use *)
   mutable nxt_z     : int;                            (* nxt_z = lowest available z value*)
   ztagcount         : float array array array;        (* ztagcount.(z).(l) = an array of counts for tags in language l for value z *)
   zlangcount        : float array array;              (* zlangcount.(z).(l) = counts in language l for value z *)
} 
;;

let read_lex_file fn = 
   let re = Re.compile (UReStr.regexp "^\\([^{White_Space}]+\\)\\(\t[.{Latin & Lu}][^{White_Space}]*\\)+$") in
   let res = Ht.create 5000 in  
   let input = new Ul.input_line (new Ce.in_channel Ce.utf16le (open_in fn)) in
   begin try 
      let ln_idx = ref 0 in
      while true do
         incr ln_idx;
         let ln = input#get () in
         let m = A.map_filter
            (function None -> None | Some t -> Some (Re.SubText.excerpt t))
            begin try
               (Opt.get (Re.regexp_match ~sem:`Longest re ln 0)) 
            with Opt.No_value ->  raise (Invalid_argument (Format.sprintf "lex file malformed: %s, line %d" fn !ln_idx)) end
         in
         assert (A.length m >= 3); 
         let w = Cm.lowercase m.(1) in
         let pos = A.map 
            (fun u -> S.strip (utf16_to_str u)) 
            (A.slice ~first:2 m) in
         if not (Ht.mem res w) then Ht.add res w (Hs.empty());
         Hs.update (Ht.find res w) (hs_of_array pos)
      done 
   with End_of_file -> () end; 
   res
;;
let read_data_file fn = 
   let re1 = Re.compile (UReStr.regexp "^O[^{White_Space}]+\t[^{White_Space}]+\t[^{White_Space}]+/[.{Latin & Lu}][^{White_Space}]*\\( [^{White_Space}]+/[.{Latin & Lu}][^{White_Space}]*\\)*$") in
   let re2 = Re.compile (UReStr.regexp "\\([^{White_Space}]+\\)/\\([.{Latin & Lu}][^{White_Space}]*\\)\\( \\|$\\)") in
   let re3 = Re.compile (UReStr.regexp "^O[^{White_Space}]+\t[^{White_Space}]+\t") in
   let sentences = ref [] in
   let input = new Ul.input_line (new Ce.in_channel Ce.utf16le (open_in fn)) in
   let counter = ref 1 in
   begin try
      while true do
         let ln = input#get () in
         let words = ref [] in         
         assert (Re.string_match re1 ln 0);
         (* Format.printf "%d\n" !counter; *)
         incr counter;
         let prefix     = Opt.get (Opt.get (Re.regexp_match ~sem:`Longest re3 ln 0)).(0) in   
         let start_idx  = 1 + Re.SubText.ur_index_of prefix (Re.SubText.last prefix) in 
         let m = ref (Re.regexp_match ~sem:`Longest re2 ln start_idx) in 
         assert (!m != None);     
         while (!m != None) do   
            let n = Opt.get !m in
            words := (Cm.lowercase(Re.SubText.excerpt(Opt.get n.(1))),utf16_to_str (Re.SubText.excerpt (Opt.get n.(2)))) :: !words;
            let st = Opt.get n.(0) in
            let idx = Re.SubText.ur_index_of st (Re.SubText.last st) in
            m := Re.regexp_match ~sem:`Longest re2 ln (idx + 1)
         done;
         sentences := aol (L.rev !words) :: !sentences;
      done
   with End_of_file -> () end;
   aol (L.rev !sentences)
;;
let read_align_file fn langs = 
   let lang_map   = Ht.of_array langs in
   let lines      = lines_of_file fn in
   let zht        = Ht.create 5000 in

   let skipped    = Hs.empty() in
   let idx        = ref 0 in
   A.switeri lines (fun i ln ->
      begin try
         let alns = splita ln in
         let alnhs = Hs.empty () in
         for a = 0 to A.length alns - 1 do
            let l, x = S.split alns.(a) "." in
            let s, w = Pair.map ios (S.split x "|") in
            if not (Ht.mem lang_map l) then 
               Hs.add skipped l
            else begin
               let l = Ht.find lang_map l in
               Hs.add alnhs (l,s,w);
            end;
         done;
         if Hs.length alnhs > 1 then begin
            let alnarr = (Hs.to_array alnhs) in
            A.sort (fun (l,s,w) (l',s',w') -> compare l l') alnarr;
            if not (Ht.mem zht alnarr) then begin
               Ht.add zht alnarr !idx;
               idx := !idx + 1;
            end;
         end;
       with 
         | Not_found -> ()
         | ExtString.Invalid_string -> 
            failwith (Format.sprintf "problem with alignment file, line = %d" (i+1)) end;
   ); 
   Hs.iter (fun l -> brintf "skipped alignments with language:   %s" l) skipped;

   (* brintf "Total latet variables = %d" (Ht.length zht); *)
   let z_values   = A.make (Ht.length zht) 0 in
   let z_locations = A.init (Ht.length zht) (fun _ -> A.make (A.length langs) (-1,-1);) in
   Ht.iter (fun alnarr idx -> A.iter (fun (l,s,w) -> z_locations.(idx).(l) <- (s,w);) alnarr; ) zht;

   (z_values, z_locations)
;;
let compile_data l lex data zht = 
   let pos_set    = Hs.empty() in
   E.iter (fun lx -> Hs.update pos_set lx) (Ht.values lex);
   let word_set   = Hs.empty() in
   A.iter_2D
      (fun (u,s) ->
         Hs.add word_set u; 
         Hs.add pos_set s;
         (* supplement lex with info from corpus *)
         if not (Ht.mem lex u) then Ht.add lex u (Hs.empty());
         Hs.add (Ht.find lex u) s)
      data;
   let pos        = Hs.to_array pos_set in
   let words      = Hs.to_array word_set in
   A.fast_sort compare pos;
   (* add dummy POS for BEGIN and END *)
   let pos        = A.init (A.length pos + 1) (fun i -> if i = 0 then "#" else pos.(i-1)) in 
   A.fast_sort compare words;
   let pos_map    = Ht.of_array pos in
   let word_map   = Ht.of_array words in
   (* compile lex' and data' -- i.e. the versions that only use integers *)
   let lex' = A.map 
      (fun w -> 
         let hs = Ht.find lex w in 
         let res = A.map (fun s -> Ht.find pos_map s) (Hs.to_array hs) in
         A.fast_sort compare res;
         res)
      words 
   in
      
   let data' = A.mapij_2D 
      (fun s w (ucd, str) -> 
         let w_id  = Ht.find word_map ucd in
         let t_id  = Ht.find pos_map str in
         let psbl  = lex'.(w_id) in
         let probs = A.make (A.length psbl) 0.0 in
         let init  = psbl.(Random.int (A.length psbl)) in
         {  id    = w_id;
            g_pos = t_id;
            p_pos = init;
            x_pos = 0;
            probs = probs;
            zindex = if Ht.mem zht (l,s,w) then Hs.to_array (Ht.find zht (l,s,w)) else [||];
         })
      data in
   (data', {words=words; pos=pos}, lex')
;;
   
let reduce_lex d n : unit =
   brintf "REDUCING LEX...";
   let counts = Ht.create 15000 in
   for l = 0 to A.length d.langs - 1 do
      Ht.clear counts;
      let pun_tag1 = A.find_item "PUN" d.maps.(l).pos in
      let pun_tag2 = A.find_item "." d.maps.(l).pos in
      let rest_tags = L.range 1 (A.length d.maps.(l).pos) |> (flip L.remove) pun_tag1 |> (flip L.remove) pun_tag2 |> aol in 
      for s = 0 to A.length d.sents.(l) - 1 do
         for w = 0 to A.length d.sents.(l).(s) - 1 do
            if not (d.lexs.(l).(d.sents.(l).(s).(w).id) = [|pun_tag1|]) && not (d.lexs.(l).(d.sents.(l).(s).(w).id) = [|pun_tag2|]) then
               Ht.incr counts d.sents.(l).(s).(w).id
         done
      done;
      let top = counts |> Ht.to_rev_sorted_array |> A.slice ~last:n |> A.map fst in
      if d.langs.(l) = "en" then
         brintf "reducing en lexicon to: %s" (A.to_str1 id (A.map (fun i -> utf16_to_str d.maps.(l).words.(i)) top));
      let top_hs = Hs.of_array top in
      for w = 0 to A.length d.lexs.(l) - 1 do
         if not (Hs.mem top_hs w) && not (d.lexs.(l).(w) = [|pun_tag1|]) && not (d.lexs.(l).(w) = [|pun_tag2|]) then
            d.lexs.(l).(w) <- rest_tags;
      done;
      (* re-assign random tags *)
      for s = 0 to A.length d.sents.(l) - 1 do
         for w = 0 to A.length d.sents.(l).(s) - 1 do
            let word = d.sents.(l).(s).(w) in
            let psbl = d.lexs.(l).(word.id) in
            let probs = A.make (A.length psbl) 0.0 in
            word.probs <- probs;
            word.p_pos <- psbl.(Random.int (A.length psbl))
         done
      done
   done
;;
let reduce_lex_freq d f : unit =
   brintf "REDUCING LEX (remove words with frequency %d or less)..." f;
   let counts = Ht.create 15000 in
   for l = 0 to A.length d.langs - 1 do
      Ht.clear counts;
      let tags = L.range 1 (A.length d.maps.(l).pos) |> aol in 
      for s = 0 to A.length d.sents.(l) - 1 do
         for w = 0 to A.length d.sents.(l).(s) - 1 do
            Ht.incr counts d.sents.(l).(s).(w).id
         done
      done;
      for w = 0 to A.length d.lexs.(l) - 1 do
         if (Ht.find counts w) <= f then
            d.lexs.(l).(w) <- tags;
      done;
      (* re-assign random tags *)
      for s = 0 to A.length d.sents.(l) - 1 do
         for w = 0 to A.length d.sents.(l).(s) - 1 do
            let word = d.sents.(l).(s).(w) in
            let psbl = d.lexs.(l).(word.id) in
            let probs = A.make (A.length psbl) 0.0 in
            word.probs <- probs;
            word.p_pos <- psbl.(Random.int (A.length psbl))
         done
      done
   done
;;
let read_data (fn : string) : data = 
   let config        = Ht.of_file id id fn in
   let langs         = splita (Ht.find config "languages") in
   let lex_files     = splita (Ht.find config "lexica") in
   let data_files    = splita (Ht.find config "data") in
   let align_file    = (splita (Ht.find config "align")).(0) in
   let testpoints    = A.map ios (splita (Ht.find config "testpoints")) in
   let suppoints     = A.map (fun s -> if s = "None" then None else Some (Pair.map ios (S.split s "--"))) (splita (Ht.find config "suppoints")) in
   let endpoints     = A.mapi (fun i s -> if s = "None" then testpoints.(i) else ios s) (splita (Ht.find config "endpoints")) in
   (* let filter        = match (splita (Ht.find config "filter")).(0) with 
                        "or" -> Some (||) | "and" -> Some (&&) | "none" -> None | _ -> failwith ("bad filter in config") in *)
   let lengths       = L.map A.length [langs; lex_files; data_files] in
   assert (L.all_same lengths);
   assert (L.hd lengths = A.length testpoints);
   let n             = L.hd lengths in
   brintf "Reading lexicon files...";
   let lexs          = A.map read_lex_file lex_files in
   brintf "Reading data files...";
   let datas         = A.map read_data_file data_files in
   let (z_values,z_locations) = read_align_file align_file langs in
   (* brintf "Checking alignments...";
   check_align align datas; *)
   (* make hash table for z_indices of each word *)
   let zht  = Ht.create 5000 in
   let numl = A.length langs in
   for zidx = 0 to A.length z_locations - 1 do
      let locs = z_locations.(zidx) in
      for l = 0 to numl - 1 do
         let (s,w) = locs.(l) in
         if (s >=0 ) && (w >= 0) then begin
            if not (Ht.mem zht (l,s,w)) then
               Ht.add zht (l,s,w) (Hs.empty ());
            Hs.add (Ht.find zht (l,s,w)) zidx;
            (* assert (zidx > 0); *)
         end;
      done;
   done;
      
   brintf "Compiling data...";
   let compiled_data = A.init n (fun i -> compile_data i lexs.(i) datas.(i) zht) in
   let res = {  
      langs = langs;
      sents = A.init n (fun i -> match compiled_data.(i) with x,_,_ -> x);
      maps  = A.init n (fun i -> match compiled_data.(i) with _,x,_ -> x);
      lexs  = A.init n (fun i -> match compiled_data.(i) with _,_,x -> x);
      n     = foi (A.length z_values);
      testpoints  = testpoints;
      suppoints   = suppoints;
      endpoints   = endpoints;
      z_values    = z_values;
      z_locations = z_locations
   } in
   let reduce = bos (splita (Ht.find config "reduce_lex")).(0) in
   let freq   = (splita (Ht.find config "frequency")).(0) in
   if reduce then if freq = "none" then reduce_lex res 100 else reduce_lex_freq res (ios freq);
   (* if reduce && filter != None then filter_align res (Opt.get filter); *)
   (* if reduce then reassign_tags res; *)
   (* brintf "d.n=%f" res.n; *)
   res 
;;
let config        = Ht.of_file id id Sys.argv.(1);;
let init_model (d : data) : model =
   let num_langs  = A.length d.langs in
   let num_z      = 1000 in
   let m = { 
      theta        = A.make num_langs (fos (splita (Ht.find config "theta_o")).(0));
      eta          = A.make num_langs (fos (splita (Ht.find config "eta_o")).(0));
      zeta         = A.make num_langs (fos (splita (Ht.find config "zeta_o")).(0));

      emcount      = A.init num_langs (fun l -> A.make_matrix (A.length d.maps.(l).pos) (A.length d.maps.(l).words) 0.);
      monocount    = A.init num_langs (fun l -> A.make (A.length d.maps.(l).pos) 0.);
      
      tagwordcount = A.init num_langs (fun l -> A.make (A.length d.maps.(l).pos) 0.);

      trcount      = A.init num_langs (fun l -> let nt = A.length d.maps.(l).pos in A.make_3D nt nt nt 0.);
      bicount      = A.init num_langs (fun l -> let nt = A.length d.maps.(l).pos in A.make_matrix nt nt 0.);

      emcountx      = A.init num_langs (fun l -> A.make_matrix (A.length d.maps.(l).pos) (A.length d.maps.(l).words) 0.);
      monocountx    = A.init num_langs (fun l -> A.make (A.length d.maps.(l).pos) 0.);
      trcountx      = A.init num_langs (fun l -> let nt = A.length d.maps.(l).pos in A.make_3D nt nt nt 0.);
      bicountx      = A.init num_langs (fun l -> let nt = A.length d.maps.(l).pos in A.make_matrix nt nt 0.);

      zcount       = A.make num_z 0.;
      ztotal       = 0.0 ;
      activez      = Hs.empty ();
      nxt_z        = 1; 
      alpha        = (fos (splita (Ht.find config "alpha")).(0));
      ztagcount    = A.init_2D num_z num_langs (fun _ l -> A.make (A.length d.maps.(l).pos) 0.0);
      zlangcount    = A.make_2D num_z num_langs 0.0;
   } in
   m
;;

let d = read_data Sys.argv.(1);;
let m = init_model d;;
let num_init_z    = ios (splita (Ht.find config "num_z")).(0);;
let start_tags    = match (splita (Ht.find config "start_tags")).(0) with "None" -> None | fn -> Some fn 

let prev_two_tags l s w =
   let sent = d.sents.(l).(s) in
   let len = A.length sent in
   try match w with
   | 0                  -> (0, 0)
   | 1                  -> (0, sent.(w-1).p_pos)
   | x when x = len + 1 -> (sent.(w-2).p_pos, 0) 
   | _                  -> (sent.(w-2).p_pos, sent.(w-1).p_pos)
   with Invalid_argument "index out of bounds" -> 
      failwith(Format.sprintf "prev_two_tags:  len = %d, w = %d" len w) 
;;
let next_two_tags l s w = 
   let sent = d.sents.(l).(s) in
   let len = A.length sent in
   try match w with
   | x when x = len - 1    -> (0, 0)
   | x when x = len - 2    -> (sent.(w+1).p_pos, 0)
   | _                     -> (sent.(w+1).p_pos, sent.(w+2).p_pos)
   with Invalid_argument "index out of bounds" -> 
      failwith(Format.sprintf "next_two_tags:  len = %d, w = %d" len w) 
;;

let init_latent () : unit =
   let posht = Ht.create 20 in
   let value = ref 0 in
   for l=0 to A.length d.langs - 1 do
      for i = 0 to A.length d.maps.(l).pos - 1 do
         let tagstr = d.maps.(l).pos.(i) in
         if not (Ht.mem posht tagstr) then begin
            Ht.add posht tagstr !value;
            value := !value + 1;
         end;
      done;
   done;
   let numz = Ht.length posht - 1 in
   (* brintf "numz=%d" numz; *)
   (* set values of z's *)
   for zindex = 0 to A.length d.z_values - 1 do
      let tcount = A.make (Ht.length posht) 0 in
      for l = 0 to A.length d.langs - 1 do
         let (s,w) = d.z_locations.(zindex).(l) in
         if (s >= 0 ) && (w >= 0) then begin
            let word   = d.sents.(l).(s).(w) in
            let tag    = word.p_pos in
            let tagstr = d.maps.(l).pos.(tag) in
            let value  = Ht.find posht tagstr in
            (* if A.length d.lexs.(l).(word.id) < 8 then *)
               A.incr tcount value;
         end;
      done;
      let maxcount = (A.max (fun a -> a;) tcount) in
      if maxcount > 1 then
         d.z_values.(zindex) <- A.find_item maxcount tcount
      else
         d.z_values.(zindex) <- (Random.int numz) + 1;
      (* assert(d.z_values.(zindex)>0); *)
   done;
;;
let init_sup () : unit =
   A.iteri (begin fun l opt -> match opt with
      | Some (a, b) ->
         for s = a to min b (A.length d.sents.(l) - 1) do
            let sent = d.sents.(l).(s) in
            for w = 0 to A.length sent - 1 do
               sent.(w).p_pos <- sent.(w).g_pos;
               sent.(w).x_pos <- sent.(w).g_pos
            done
         done
      | None -> ()
   end) d.suppoints
;;
let init_counts () : unit =
   (* updates counts while iterating through data *)
   for l = 0 to A.length d.langs - 1 do
      for s = 0 to d.endpoints.(l) do
         let sen = d.sents.(l).(s) in
         let slen = A.length sen - 1 in
         for w = 0 to slen do

            let word = sen.(w) in
            let prevprevtag, prevtag = prev_two_tags l s w in
            let tag = word.p_pos in
            A.incr' m.emcount.(l).(tag) word.id ;
            A.incr' m.monocount.(l) tag ;
            A.incr' m.trcount.(l).(prevprevtag).(prevtag) tag;
            A.incr' m.bicount.(l).(prevprevtag) prevtag;

            if w == slen then begin
               A.incr' m.trcount.(l).(prevtag).(tag) 0 ;
               A.incr' m.trcount.(l).(tag).(0) 0 ;
               A.incr' m.bicount.(l).(prevtag) tag ;
               A.incr' m.bicount.(l).(tag) 0 ;
            end;

            for idx = 0 to A.length word.zindex - 1 do
               let z = d.z_values.(word.zindex.(idx)) in
               A.incr' m.zlangcount.(z) l;
               A.incr' m.ztagcount.(z).(l) tag;
            done;
            
         done
      done;
      for w = 0 to A.length d.lexs.(l) - 1 do
         for t = 0 to A.length d.lexs.(l).(w) -1 do
            A.incr' m.tagwordcount.(l) d.lexs.(l).(w).(t)
         done;
      done;
   done;
;;
let init_latent_counts () : unit =
   for zindex = 0 to A.length d.z_values - 1 do
      let z = d.z_values.(zindex) in
      let in_train = ref true in
      for l = 0 to A.length d.langs - 1 do
         let (s,w) = d.z_locations.(zindex).(l) in
         if s > d.endpoints.(l) then
            in_train := false;
      done;
      if !in_train then begin
         if z >= m.nxt_z then
            m.nxt_z <- z + 1;
         Hs.add m.activez z;
         A.incr' m.zcount z;
         m.ztotal <- m.ztotal +. 1.0;
      end;
   done;
;;
let init_tags fn = 
   print_endline ("Initalizing tags according to: " ^ fn);
   L.iter 
      (begin fun ln ->
         begin try
            let (l, s, w, t) = match split ln with
               | [l; s; w; t]  -> 
                  let l' = A.find_item l d.langs in
                  let s' = ios s in
                  let w' = ios w in
                  let t' = if t = "*" then A.find_item "PUN" d.maps.(l').pos else A.find_item t d.maps.(l').pos in 
                  (l', s', w', t')
               | _ -> failwith ("malformed init_tags file: *" ^ ln ^ "*")
            in d.sents.(l).(s).(w).p_pos <- t;
         with 
            | Not_found -> ()
         end;
      end)
      (list_of_file fn)
;;


Opt.may init_tags start_tags;;
init_sup ();;
init_latent ();;
init_counts ();;
init_latent_counts ();;

let final_stop = 1000;;
let r = ref 0;;

(*************************************************************************************)
let is_sup l s = 
   match d.suppoints.(l) with
      | None ->      false
      | Some (a, b) ->  s >= a && s <= b
;;
let prev_two_xtags l s w =
   let sent = d.sents.(l).(s) in
   let len = A.length sent in
   try match w with
      | 0                  -> (0, 0)
      | 1                  -> (0, sent.(w-1).x_pos)
      | x when x = len + 1 -> (sent.(w-2).x_pos, 0)
      | _                  -> (sent.(w-2).x_pos,
   sent.(w-1).x_pos)
   with Invalid_argument "index out of bounds" ->
      failwith(Format.sprintf "prev_two_max_tags:  len = %d, w = %d" len w)
;;
let next_two_xtags l s w =
   let sent = d.sents.(l).(s) in
   let len = A.length sent in
   try match w with
      | x when x = len - 1    -> (0, 0)
      | x when x = len - 2    -> (sent.(w+1).x_pos, 0)
      | _                     -> (sent.(w+1).x_pos,
   sent.(w+2).x_pos)
   with Invalid_argument "index out of bounds" ->
      failwith(Format.sprintf "next_two_max_tags:  len = %d, w = %d" len w)
;;
let choose_xtags () =
   let num_langs = A.length d.langs in
   for l = 0 to num_langs - 1 do
      for s = 0 to d.endpoints.(l) - 1 do
         let sen = d.sents.(l).(s) in
         for w = 0 to A.length sen - 1 do
            let word  = sen.(w) in
            let probs = word.probs in
            let maxi  = ref 0 in
            for i = 1 to A.length probs - 1 do
               if probs.(i) > probs.(!maxi) then
                  maxi := i;
            done;
            word.x_pos <- d.lexs.(l).(word.id).(!maxi);
         done;
      done;
   done;
;;
let update_xcounts () : unit =
   (* initially set all counts to 0 *)
   for l = 0 to A.length d.langs - 1 do
      A.fill_all_matrix m.bicountx.(l) 0.0 ;
      A.fill_all_matrix m.emcountx.(l) 0.0 ;
      for t = 0 to A.length d.maps.(l).pos - 1 do
         A.fill_all_matrix m.trcountx.(l).(t) 0.0 ;
      done;
      A.fill_all m.monocountx.(l) 0.0;
   done;
   (* updates counts while iterating through data *)
   for l = 0 to A.length d.langs - 1 do
      for s = 0 to d.endpoints.(l) do
         let sen = d.sents.(l).(s) in
         let slen = A.length sen - 1 in
         A.incr' m.bicountx.(l).(0) 0 ;
         for w = 0 to slen do
            let word = sen.(w) in
            let prevprevtag, prevtag = prev_two_xtags l s w in
            let tag = word.x_pos in
            A.incr' m.trcountx.(l).(prevprevtag).(prevtag) tag ;
            A.incr' m.emcountx.(l).(tag) word.id ;
            A.incr' m.monocountx.(l) tag ;
            if w == slen then begin
               A.incr' m.trcountx.(l).(prevtag).(tag) 0 ;
               A.incr' m.trcountx.(l).(tag).(0) 0 ;
            end;    
            A.incr' m.bicountx.(l).(prevtag) tag ;
         done;
      done;
   done;
;;

let viterbi l s x : unit =
 (* sentence level viterbi  *)
  let sen = d.sents.(l).(s) in
  let size = A.length sen in
  let tagset_size = A.length d.maps.(l).pos in
  let pmat = A.make_matrix tagset_size tagset_size neg_infinity in
  let bmat = A.make_matrix tagset_size tagset_size 0 in
  
   let (trcount,bicount,emcount,monocount) =
      if x then (m.trcountx, m.bicountx, m.emcountx, m.monocountx)
      else (m.trcount, m.bicount, m.emcount, m.monocount) in

  let prob_matrix_arr = A.init (size + 1) (fun _ -> A.copy_matrix pmat) in
  let back_matrix_arr = A.init (size + 1) (fun _ -> A.copy_matrix bmat) in

   prob_matrix_arr.(0).(0).(0) <- 0.0;
   back_matrix_arr.(0).(0).(0) <-  -1;
  
   let alpha   = m.eta.(l) in
   let beta    = m.theta.(l) in
   let t_alpha = alpha *. float tagset_size in
   for w = 1 to size do
      let word = sen.(w-1).id in
      let wtags = d.lexs.(l).(word) in
      for i = 0 to A.length wtags - 1 do
         let tag = wtags.(i) in
         for tag' = 0 to tagset_size - 1 do
            let maxp = ref neg_infinity in
            for tag'' = 0 to tagset_size - 1 do
               let p = prob_matrix_arr.(w-1).(tag'').(tag') +.
                  log10 ( ( trcount.(l).(tag'').(tag').(tag) +. alpha ) /. ( bicount.(l).(tag'').(tag') +. t_alpha ) ) +.
                  log10 ( ( emcount.(l).(tag).(word) +. beta ) /. ( monocount.(l).(tag) +. beta *.  m.tagwordcount.(l).(tag) ) ) in
               if p > !maxp then begin
                  maxp := p ;
                  prob_matrix_arr.(w).(tag').(tag) <- p ;
                  back_matrix_arr.(w).(tag').(tag) <- tag''
               end
            done
         done
      done
   done;

  let thistag = ref 0 in
  let prevtag = ref 0 in
  let max = ref (log10 0.0) in
  for tag' = 0 to tagset_size - 1 do
    for tag = 0 to tagset_size - 1 do
       let p = prob_matrix_arr.(size).(tag').(tag) +. 
            log10 ( ( trcount.(l).(tag').(tag).(0) +.  alpha) /. ( bicount.(l).(tag').(tag) +. t_alpha ) )in
       if p > !max then begin
          thistag := tag ;
          prevtag := tag';
          max := p ;
       end
    done;
   done;

   for w = size downto 1 do

      if x then 
         sen.(w-1).x_pos <- !thistag
      else
         sen.(w-1).p_pos <- !thistag;

     let tempthistag = !thistag in
     thistag := !prevtag;
     if w > 1 then   
        prevtag := back_matrix_arr.(w).(!prevtag).(tempthistag)
   done
;;
let annotate_testdata x : unit =
   let langcount = A.length d.langs in
   if x then
      update_xcounts ();
   for l = 0 to langcount - 1 do
      for s = d.testpoints.(l) + 1 to A.length d.sents.(l) - 1 do
         viterbi l s x
      done;
   done;  
;;

let compute_word_prob l s w : float =
   let word = d.sents.(l).(s).(w) in
   let tag  = word.p_pos in
   ( ( m.emcount.(l).(tag).(word.id) +. m.theta.(l) ) /.
     ( m.monocount.(l).(tag) +. m.tagwordcount.(l).(tag) *. m.theta.(l) ) ) 
;; 
let compute_tag_prob l s w =
   let words = d.sents.(l).(s) in
   let (tag, zindex) = match w with
      | x when x > A.length d.sents.(l).(s) - 1 ->   (0, [|0|])
      | _ ->  (words.(w).p_pos, words.(w).zindex) in
   let pp_tag, p_tag = prev_two_tags l s w in
   let prob = ref 1. in
   let tsetsize = foi (A.length d.maps.(l).pos) in
   assert (m.trcount.(l).(pp_tag).(p_tag).(tag) >= 0.);
   assert (m.bicount.(l).(pp_tag).(p_tag) >= 0.);
   prob := ( ( m.trcount.(l).(pp_tag).(p_tag).(tag) +. m.eta.(l) ) /.
      ( m.bicount.(l).(pp_tag).(p_tag) +. tsetsize  *. m.eta.(l) ) );
   for idx = 0 to A.length zindex - 1 do
      let z =  d.z_values.(zindex.(idx)) in
      prob := !prob *. ( ( m.ztagcount.(z).(l).(tag) +. m.zeta.(l) ) /.
         ( m.zlangcount.(z).(l) +. tsetsize  *. m.zeta.(l) ) );
      assert (!prob > 0.);
   done;
   let prttn = ref 0. in
   for tag' = 0 to A.length d.maps.(l).pos - 1 do
      let score = ref 1. in
      score := ( ( m.trcount.(l).(pp_tag).(p_tag).(tag') +. m.eta.(l) ) /.
         ( m.bicount.(l).(pp_tag).(p_tag) +. tsetsize  *. m.eta.(l) ) );
      for idx = 0 to A.length zindex - 1 do
         let z =  d.z_values.(zindex.(idx)) in
         assert (m.trcount.(l).(pp_tag).(p_tag).(tag') >= 0.);
         score := !score *. ( ( m.ztagcount.(z).(l).(tag') +. m.zeta.(l) ) /.
         ( m.zlangcount.(z).(l) +. tsetsize  *. m.zeta.(l) ) );
      done;
      assert (!score > 0.);
      prttn := !prttn +. !score;
   done;
   assert (!prob > 0.);
   assert (!prttn > 0.);
   !prob /. !prttn
;;

let incr_tag_counts l s w : unit = 
   let thistag = d.sents.(l).(s).(w).p_pos in
   let pptag, ptag = prev_two_tags l s w in
   let ntag, nntag = next_two_tags l s w in
   let sent = d.sents.(l).(s) in
   
   for idx = 0 to A.length sent.(w).zindex - 1 do
      let z =  d.z_values.(sent.(w).zindex.(idx)) in
      A.incr' m.ztagcount.(z).(l) thistag ;
   done;
   
   A.incr' m.trcount.(l).(pptag).(ptag) thistag ;
   A.incr' m.trcount.(l).(ptag).(thistag) ntag ;
   A.incr' m.bicount.(l).(ptag) thistag ;
   A.incr' m.trcount.(l).(thistag).(ntag) nntag ;
   A.incr' m.bicount.(l).(thistag) ntag ;

   A.incr' m.emcount.(l).(thistag) sent.(w).id ;
   A.incr' m.monocount.(l) thistag ;
;;
let decr_tag_counts l s w : unit =
   let thistag = d.sents.(l).(s).(w).p_pos in
   let pptag, ptag = prev_two_tags l s w in
   let ntag, nntag = next_two_tags l s w in
   let sent = d.sents.(l).(s) in
   
   for idx = 0 to A.length sent.(w).zindex - 1 do
      let z =  d.z_values.(sent.(w).zindex.(idx)) in
      A.decr' m.ztagcount.(z).(l) thistag ;
   done;
   
   A.decr' m.trcount.(l).(pptag).(ptag) thistag ;
   A.decr' m.trcount.(l).(ptag).(thistag) ntag ;
   A.decr' m.bicount.(l).(ptag) thistag ;
   A.decr' m.trcount.(l).(thistag).(ntag) nntag ;
   A.decr' m.bicount.(l).(thistag) ntag ;

   A.decr' m.emcount.(l).(thistag) sent.(w).id ;
   A.decr' m.monocount.(l) thistag ;
;;
let sample_tag l s w : unit = 
   assert (not (is_sup l s));
   let word = d.sents.(l).(s).(w) in
   let lex  = d.lexs.(l).(word.id) in
   let len = A.length lex in
   let p = A.make len 0.0 in

   decr_tag_counts l s w;

   (* aggregate probabilities *)
   for i = 0 to len - 1 do
      let tag = lex.(i) in
      word.p_pos <- tag;
      p.(i) <- compute_word_prob l s w
            *. compute_tag_prob l s w
            *. compute_tag_prob l s (w+1)
            *. compute_tag_prob l s (w+2);
      if !r >= (final_stop - 100) then
         word.probs.(i) <- word.probs.(i) +. p.(i);
   done;

   (* Sample and set the tag *)
   let i = Gsl_randist.discrete rng (Gsl_randist.discrete_preproc p) in
   word.p_pos <- lex.(i);

   incr_tag_counts l s w;
;;

let incr_z_counts zindex : unit =
   let z = d.z_values.(zindex) in
   A.incr' m.zcount z;
   if m.zcount.(z) = 1. then
      Hs.add m.activez z;
   for l = 0 to A.length d.langs - 1 do
      let (s,w) = d.z_locations.(zindex).(l) in
      if s>=0 && w>=0 then begin
         let tag = d.sents.(l).(s).(w).p_pos in
         A.incr' m.ztagcount.(z).(l) tag;
         A.incr' m.zlangcount.(z) l;
      end;
   done;
;;
let decr_z_counts zindex: unit =
   let z = d.z_values.(zindex) in
   A.decr' m.zcount z;
   if m.zcount.(z) = 0. then
      Hs.remove m.activez z;
   for l = 0 to A.length d.langs - 1 do
      let (s,w) = d.z_locations.(zindex).(l) in
      if s>=0 && w>=0 then begin
         let tag = d.sents.(l).(s).(w).p_pos in
         assert (m.ztagcount.(z).(l).(tag) > 0.);
         A.decr' m.ztagcount.(z).(l) tag;
         A.decr' m.zlangcount.(z) l;
      end;
   done;
;;
let sample_latent zindex : unit =

   decr_z_counts zindex;

   let z_arr   = Hs.to_array m.activez in
   let z_n     = m.ztotal in
   let alpha   = m.alpha in   
   let p       = A.init (A.length z_arr + 1)
      begin function
         | 0 ->
            let prob = ref (alpha /. ((z_n -. 1.) +. alpha)) in
            for l = 0 to A.length d.langs - 1 do
               let (s,w)    = d.z_locations.(zindex).(l) in 
               let tsetsize = foi (A.length d.maps.(l).pos) in
               if s>=0 && w>=0 then
                  prob := !prob *. m.zeta.(l) /. ( tsetsize *. m.zeta.(l) )
            done;
            !prob
         | i ->
              let z    = z_arr.(i-1) in
              let prob = ref (m.zcount.(z) /. (z_n -. 1. +. alpha)) in
              for l = 0 to A.length d.langs - 1 do
                 let (s,w)    = d.z_locations.(zindex).(l) in 
                 if s>=0 && w>=0 then
                    let tag = d.sents.(l).(s).(w).p_pos in
                    let tsetsize = foi (A.length d.maps.(l).pos) in
                    prob := !prob *. ( ( m.ztagcount.(z).(l).(tag) +. m.zeta.(l) ) /.
                     ( m.zlangcount.(z).(l) +. tsetsize  *. m.zeta.(l) ) );
                  done;
              !prob
      end in

   assert (A.for_all ((<) 0.) p);
   let newz = match Gsl_randist.discrete rng (Gsl_randist.discrete_preproc p) with
      | 0 -> m.nxt_z
      | i -> z_arr.(i-1)
   in
   
   d.z_values.(zindex) <- newz;
   if newz = m.nxt_z then
      m.nxt_z <- A.find_item 0. m.zcount;
         
   incr_z_counts zindex;

;;

let accept p p' q q' : bool = 
   (*assuming both likelihood and prob are in natural log *)
   (* p = old prob; p' = new prob; q = old proposal; q' = new proposal *)
  Gsl_rng.uniform rng <= exp (( p' +. q ) -. ( p +. q'))
;;
let asc_fact par count =
   let fact = ref 0.0 in
   for i = 0 to iof (count) - 1 do
      fact := !fact +. log(par +. foi i)
   done;
   !fact
;;
let estimate_theta l : unit =
   let oldtheta = m.theta.(l) in
   let oldmean  = oldtheta in
   let oldvar   = oldmean *. 0.1 in
   let newtheta = oldmean +. Gsl_randist.gaussian rng oldvar in
   let newmean  = newtheta in
   let newvar   = newmean *. 0.1 in
   let q        = log(Gsl_randist.gaussian_pdf (oldtheta -. newmean) newvar) in
   let q'       = log(Gsl_randist.gaussian_pdf (newtheta -. oldmean) oldvar) in
 
   let p      = ref 0.0 in
   let p'     = ref 0.0 in
   let numo   = ref 0.0 in
   let numn   = ref 0.0 in
   let denomo = ref 0.0 in
   let denomn = ref 0.0 in
   for t = 0 to A.length d.maps.(l).pos - 1 do
      A.iter (fun count -> numo := !numo +. asc_fact oldtheta count ) m.emcount.(l).(t);
      A.iter (fun count -> numn := !numn +. asc_fact newtheta count ) m.emcount.(l).(t);
      denomo := asc_fact (m.tagwordcount.(l).(t) *. oldtheta) m.monocount.(l).(t);
      denomn := asc_fact (m.tagwordcount.(l).(t) *. newtheta) m.monocount.(l).(t);
      p      := !p +. ( !numo -. !denomo );
      p'     := !p' +. ( !numn -. !denomn );
      numo := 0.; numn := 0.; denomo := 0.; denomn := 0.;
   done;
 
   if accept !p !p' q q' then
      m.theta.(l) <- newtheta;
;;
let estimate_eta l : unit =
   let num_tags   = A.length d.maps.(l).pos in
   let oldeta     = m.eta.(l) in
   let oldmean    = oldeta in  (* XXX assumes all etas are the same for language l*)
   let oldvar     = oldmean *. 0.1 in
   let neweta     = oldmean +. Gsl_randist.gaussian rng oldvar in
   let newmean    = neweta in
   let newvar     = newmean *. 0.1 in
   let q          = log (Gsl_randist.gaussian_pdf (oldmean -. newmean) newvar) in
   let q'         = log (Gsl_randist.gaussian_pdf (newmean -. oldmean) oldvar) in
   let p          = ref 0.0 in
   let p'         = ref 0.0 in
   let numo   = ref 0.0 in
   let numn   = ref 0.0 in
   let denomo = ref 0.0 in
   let denomn = ref 0.0 in
      
   for t = 0 to num_tags - 1 do
      for t' = 0 to num_tags - 1 do
         if (t = 0 || t' > 0) then begin   (* skip case of (n,0,..) (for n != 0) *)
            A.iter (fun count -> numo := !numo +. asc_fact oldeta count) m.trcount.(l).(t).(t');
            A.iter (fun count -> numn := !numn +. asc_fact oldeta count) m.trcount.(l).(t).(t');
            denomo := asc_fact (foi num_tags *. oldeta) m.bicount.(l).(t).(t');
            denomn := asc_fact (foi num_tags *. neweta) m.bicount.(l).(t).(t');
            p      := !p +. ( !numo -. !denomo );
            p'     := !p' +. ( !numn -. !denomn );
            numo := 0.; numn := 0.; denomo := 0.; denomn := 0.;
         end
      done
   done;
 
   if accept !p !p' q q' then
      m.eta.(l) <- neweta;
;;

let print_z () : unit =
   Hs.iter begin fun zval ->  
      brintf "z=%d" zval;
      for l = 0 to A.length d.langs - 1 do
         Printf.printf "%s " d.langs.(l);
         let scounts = A.init (A.length m.ztagcount.(zval).(l)) (fun i -> (m.ztagcount.(zval).(l).(i),d.maps.(l).pos.(i)); ) in
         A.fast_sort (fun a b -> -1 * compare (fst a) (fst b);) scounts;
         A.iter (fun (count,tag) -> Printf.printf "%5s=%.2f "  tag count; ) scounts;
         brintf "";
      done;
      brintf "";
   end m.activez;
;;

let test ?fn (final : bool) (x : bool) =
   (* returns an array of accuracy values, one for each language *)
   if x then
      choose_xtags ();
   if final then
      annotate_testdata x;
   let save = fn != None in
   let langcount = A.length d.langs in
   let res = A.make langcount 0.0 in
   for l = 0 to langcount - 1 do
      let out_str = Buffer.create (if save then 10 else 50000) in
      let good  = ref 0.0 in
      let total = ref 0.0 in
      let pun   = A.find_item "PUN" d.maps.(l).pos in
      let sm    = A.find_item "." d.maps.(l).pos in
      let start, stop = match final with
         | true   -> d.testpoints.(l) + 1, A.length d.sents.(l) -1
         | false  -> 0, d.endpoints.(l) in
      for s = start to stop do
         for w = 0 to A.length d.sents.(l).(s) - 1 do
            let p_pos = match x with
               | true   -> d.sents.(l).(s).(w).x_pos
               | false  -> d.sents.(l).(s).(w).p_pos in
            if not ( p_pos = pun || p_pos = sm ) then begin
               total := !total +. 1. ;
               if ( p_pos == d.sents.(l).(s).(w).g_pos ) then
                  good := !good +. 1. ;
            end;
            if save then
               Buffer.add_string out_str (d.maps.(l).pos.(p_pos) ^ " ");
         done;
         if save then
            Buffer.add_string out_str "\n";
      done;
      res.(l) <- !good /. !total *. 100.0;
      if save then
         output_file ~filename:(Opt.get fn).(l) ~text:(Buffer.contents out_str);
   done;
   res
;;
let sample r : unit =
   brintf "SAMPLING ROUND = %d    @%d" r (iof (Unix.time ()) - start_time);
   for i = 1 to 1 do
      for l = 0 to A.length d.langs - 1 do
         let sents = d.sents.(l) in
         for s = 0 to d.endpoints.(l) do
            if not (is_sup l s) then begin
               let sent = sents.(s) in
               for w = 0 to A.length sent - 1 do
                  sample_tag l s w
               done
            end
         done
      done;
   done;
   for idx = 0 to A.length d.z_values - 1 do
      let samplez = ref true in
      for l = 0 to A.length d.langs - 1 do
         let (s,w) = d.z_locations.(idx).(l) in
         if s > d.endpoints.(l) then
            samplez := false;
      done;
      if !samplez then
         sample_latent idx;
   done;
   
   (* hyperparameter re-estimation *)
   if r mod 1 = 0 then begin 
      for l = 0 to A.length d.langs - 1 do
         estimate_theta l;
         estimate_eta l;
      done;
   end;
   
   (* Hs.iter (fun a -> brintf "z=%d (%f)"  a m.zcount.(a);) m.activez; *)
;;
let main () = 

   (* print_z ();  *)
 
   let res = test false false in
   A.iteri (fun l r -> brintf "INITIAL accuracy for %s = %f" d.langs.(l) r) res;   

   for r = 0 to 200 do
      for l = 0 to A.length d.langs - 1 do 
         estimate_theta l; 
         estimate_eta l;
      done
   done;
   
   while (!r < final_stop) do
      sample !r;

      (* if !r mod 100 = 0 then begin
         let restrain   = test false false in
         let restrainx  = test false true in
         let restest    = test true false in
         let restestx   = test true true in
         A.iteri begin 
            fun l lang -> brintf "      %s     train(x) = %.2f %.2f test(x) = %.2f %.2f"
            lang restrain.(l) restrainx.(l) restest.(l) restestx.(l)
         end d.langs;       
      end  *)
      (* else begin
         let restrain = test false false in
         A.iteri (fun l _ -> brintf "      accuracy for %s     train = %f" d.langs.(l) restrain.(l)) d.langs ;
      end;  *)

      if !r mod 5 = 0 then begin
         (* print_endline "compacting GC";
         let tm = Unix.gettimeofday () in *)
         Gc.compact ();
         (* let diff = Unix.gettimeofday () -. tm in
         brintf "done (%.4f)" diff;
         print_z ();   *)
      end;

      incr r      
   done;
   
   let fn = A.map (fun lang -> Sys.argv.(2) ^ "-" ^  lang) d.langs in
   let restrain   = test false false in
   let restrainx  = test false true in
   let restest    = test true false in
   let restestx   = test ~fn true true in
   A.iteri begin 
      fun l lang -> brintf "      %s     train(x) = %.2f %.2f test(x) = %.2f %.2f" lang restrain.(l) restrainx.(l) restest.(l) restestx.(l)
   end d.langs;
   A.iteri (fun i l -> brintf "lang %s :   theta = %f  eta = %f" l m.theta.(i) m.eta.(i)) d.langs
;;

main ()































