
open Util
open Std
open CamomileLibrary.Default.Camomile 
module Cm = CaseMap.Make (UTF16)
module Ul = ULine.Make (UTF16)
module Ce = CharEncoding
module Re = UReStr.Make (UTF16)

let rng = Gsl_rng.make (Gsl_rng.default ());;
Random.self_init ();;
let seed = Random.nativeint (Nativeint.of_int max_int);;
Gsl_rng.set rng seed;;
let set_seed n =
  Gsl_rng.set rng (Nativeint.of_int n);
  Random.init n
;;
(* set_seed 0;; (* uncomment for debugging *) *)
let print_utf16 u = 
   let output = new Ce.out_channel Ce.utf16le stdout in
   UTF16.iter output#put u;
   output#flush ()
;;  
let utf16_of_str s = UTF16.init (S.length s) (fun i -> UChar.of_char s.[i]);;
let utf16_to_str u = S.init (UTF16.length u) (fun i -> UChar.char_of (UTF16.get u i));;
let get_get_schedule b e =
   (fun burn rnds ->
      let n = rnds - burn in
      let d = (e -. b) /. (foi n) in 
      A.init rnds
         (fun i ->
            if i < n then
               d *. (foi i) +. b
            else 1.))
;;

type alignment = { l : int; s : int; w : int }
type omega     = float array
type mapping   = { words : UTF16.t array; pos : string array}
type word      = {
   id                : int;
   g_pos             : int;
   mutable p_pos     : int;
   mutable x_pos     : int;
   mutable probs     : float array;
   mutable align     : alignment array }
;;                   
type data      = {
   langs             : string array;
   sents             : word array array array;         (* sentences.(i).(j) = language i, sentence j, array of words *)
   maps              : mapping array;                  (* mapping.(i) = language i *)
   lexs              : int array array array;          (* lexs.(i).(j) = language i, word j : array of POS's *) 
   mutable n         : float;                          (* number of alignments - (made mutable to later restrict to alignments in traing data -- tahira) *) 
   testpoints        : int array;
   endpoints         : int array;
   suppoints         : (int * int) option array;
   anneal            : bool }
;;   
type model     = {
   theta             : float array;                    (* theta.(i) = hyper param for emissions of language i *) 
   eta               : float array;                    (* eta.(l) = hyperparameters for transition dirichlet prior *)
   mutable zeta      : float;                           (* hyperparameter for coupling distribution dirichlet prior *)

   emcount           : float array array array;        (* emcount.(l).(t) = array of word counts for language l, tag t *)  
   trcount           : float array array array array;  (* trcount.(l).(t1).(t2) = array of subsequent tag counts *)  
   bicount           : float array array array;        (* bicount.(l).(t) = array of subsequent tag counts *) 
   monocount         : float array array;              (* monocount.(l) = array of tag counts *) 

   emcountx          : float array array array;           (* similar counts, but for x_pos instead of p_pos *)
   trcountx          : float array array array array; 
   bicountx          : float array array array;       
   monocountx        : float array array;             

   cpcount           : float array array;
   tagwordcount      : float array array }              (* tagwordcount.(l).(t) = number of words that can with tag t*) 
;;

let cmp_aligns a1 a2 = 
   let r1 = compare a1.l a2.l in
   let r2 = compare a1.s a2.s in
   let r3 = compare a1.w a2.w in
   if r1 = 0 then (if r2 = 0 then r3 else r2) else r1
;;
let read_lex_file fn = 
   let re = Re.compile (UReStr.regexp "^\\([^{White_Space}]+\\)\\(\t[.{Latin & Lu}][^{White_Space}]*\\)+$") in
   let res = Ht.create 5000 in  
   let input = new Ul.input_line (new Ce.in_channel Ce.utf16le (open_in fn)) in
   begin try 
      let ln_idx = ref 0 in
      while true do
         incr ln_idx;
         let ln = input#get () in
         let m = A.map_filter
            (function None -> None | Some t -> Some (Re.SubText.excerpt t))
            begin try
               (Opt.get (Re.regexp_match ~sem:`Longest re ln 0)) 
	    with Opt.No_value ->  raise (Invalid_argument (Format.sprintf "lex file malformed: %s, line %d" fn !ln_idx)) end
         in
         assert (A.length m >= 3); 
         let w = Cm.lowercase m.(1) in
         let pos = A.map 
            (fun u -> S.strip (utf16_to_str u)) 
            (A.slice ~first:2 m) in
         if not (Ht.mem res w) then Ht.add res w (Hs.empty());
         Hs.update (Ht.find res w) (hs_of_array pos)
      done 
   with End_of_file -> () end; 
   res
;;
let read_data_file fn = 
   let re1 = Re.compile (UReStr.regexp "^O[^{White_Space}]+\t[^{White_Space}]+\t[^{White_Space}]+/[.{Latin & Lu}][^{White_Space}]*\\( [^{White_Space}]+/[.{Latin & Lu}][^{White_Space}]*\\)*$") in
   let re2 = Re.compile (UReStr.regexp "\\([^{White_Space}]+\\)/\\([.{Latin & Lu}][^{White_Space}]*\\)\\( \\|$\\)") in
   let re3 = Re.compile (UReStr.regexp "^O[^{White_Space}]+\t[^{White_Space}]+\t") in
   let sentences = ref [] in
   let input = new Ul.input_line (new Ce.in_channel Ce.utf16le (open_in fn)) in
   let counter = ref 1 in
   begin try
      while true do
         let ln = input#get () in
         let words = ref [] in
         assert (Re.string_match re1 ln 0);
         incr counter;
         let prefix     = Opt.get (Opt.get (Re.regexp_match ~sem:`Longest re3 ln 0)).(0) in   
         let start_idx  = 1 + Re.SubText.ur_index_of prefix (Re.SubText.last prefix) in 
         let m = ref (Re.regexp_match ~sem:`Longest re2 ln start_idx) in 
         assert (!m != None);     
         while (!m != None) do   
            let n = Opt.get !m in
            words := (Cm.lowercase(Re.SubText.excerpt(Opt.get n.(1))),utf16_to_str (Re.SubText.excerpt (Opt.get n.(2)))) :: !words;
            let st = Opt.get n.(0) in
            let idx = Re.SubText.ur_index_of st (Re.SubText.last st) in
            m := Re.regexp_match ~sem:`Longest re2 ln (idx + 1)
         done;
         sentences := aol (L.rev !words) :: !sentences;
      done
   with End_of_file -> () end;
   aol (L.rev !sentences)
;;
let read_align_file fn langs = 
   let res        = Ht.create 5000 in
   let lang_map   = Ht.of_array langs in
   let lines      = lines_of_file fn in
   let skipped    = Hs.empty() in
   A.switeri lines (fun i ln ->
      begin try 
         let a, b = S.split ln "-" in
         let l1, a' = S.split a "." in
         let l2, b' = S.split b "." in
         let s1, w1 = Pair.map ios (S.split a' "|") in
         let s2, w2 = Pair.map ios (S.split b' "|") in 
         if not (Ht.mem lang_map l1) then 
            Hs.add skipped l1;
         if not (Ht.mem lang_map l2) then
            Hs.add skipped l2;
         let l1 = Ht.find lang_map l1 in
         let l2 = Ht.find lang_map l2 in
         let align1 = {l = l1;  s = s1;  w = w1} in
         let align2 = {l = l2;  s = s2;  w = w2} in
         if not (Ht.mem res align1) then
            Ht.add res align1 (Hs.empty());
         if not (Ht.mem res align2) then
            Ht.add res align2 (Hs.empty());
         Hs.add (Ht.find res align1) align2;
         Hs.add (Ht.find res align2) align1
      with 
         | Not_found -> ()
         | ExtString.Invalid_string -> 
            failwith (Format.sprintf "problem with alignment file, line = %d" (i+1)) end;
   ); 
   
   Hs.iter (fun l -> brintf "skipped alignments with language:   %s" l) skipped;
   res
;;
let remove_crossing_alignments aligns = 
   let align_keys0 = Ht.keys aligns |> E.filter (fun a -> a.l = 0) |> aoe in
   A.fast_sort cmp_aligns align_keys0;
   
   let prev_s = ref (-1) in
   let prev_w = ref (-1) in
   
   A.iter (begin fun a ->
      assert (a.l = 0);
      assert (Hs.length (Ht.find aligns a) = 1);
      let a' = (Hs.to_array (Ht.find aligns a)).(0) in
      let l', s', w' = a'.l, a'.s, a'.w in
      assert (l' = 1);
      if s' > !prev_s || (s' = !prev_s && w' > !prev_w) then begin
         prev_s := s';
         prev_w := w';
      end else begin       (* crossing edge! *)
         Ht.remove aligns a;
         Ht.remove aligns a'
      end
   end) align_keys0
;;
let compile_data l lex data align = 
   let pos_set    = Hs.empty() in
   E.iter (fun lx -> Hs.update pos_set lx) (Ht.values lex);
   let word_set   = Hs.empty() in
   A.iter_2D
      (fun (u,s) ->
         Hs.add word_set u; 
         Hs.add pos_set s;
         (* supplement lex with info from corpus *)
         if not (Ht.mem lex u) then Ht.add lex u (Hs.empty());
         Hs.add (Ht.find lex u) s)
      data;
   let pos        = Hs.to_array pos_set in
   let words      = Hs.to_array word_set in
   A.fast_sort compare pos;
   (* add dummy POS for BEGIN and END *)
   let pos        = A.init (A.length pos + 1) (fun i -> if i = 0 then "#" else pos.(i-1)) in 
   A.fast_sort compare words;
   let pos_map    = Ht.of_array pos in
   let word_map   = Ht.of_array words in
   (* compile lex' and data' -- i.e. the versions that only use integers *)
   let lex' = A.map 
      (fun w -> 
         let hs = Ht.find lex w in 
         let res = A.map (fun s -> Ht.find pos_map s) (Hs.to_array hs) in
         A.fast_sort compare res;
         res)
      words 
   in
   let data' = A.mapij_2D 
      (fun s w (ucd, str) -> 
         let w_id = Ht.find word_map ucd in
         let t_id = Ht.find pos_map str in
         let psbl = lex'.(w_id) in
         let probs = A.make (A.length psbl) 0.0 in
         let init = psbl.(Random.int (A.length psbl)) in
         let aln  = Hs.to_array (Ht.get align {l=l;s=s;w=w} (Hs.empty())) in
         assert (A.length aln = 0 || A.length aln = 1);
         {  id    = w_id;
            g_pos = t_id;
            p_pos = init;
            x_pos = 0;
            probs = probs;
            align = Hs.to_array (Ht.get align {l=l; s=s; w=w} (Hs.empty()))
         })
      data in
   (data', {words=words; pos=pos}, lex')
;;
let check_align aligns datas langs = 
   (* checks to ensure the indices in [align] are consistent with [datas] *)
   let align_keys = Ht.keys aligns in
   
   E.iter (begin fun align ->
      assert (Hs.length (Ht.find aligns align) = 1);
      let align' = (Hs.to_array (Ht.find aligns align)).(0) in 
      assert (Hs.length (Ht.find aligns align') = 1);
      let back_align = (Hs.to_array (Ht.find aligns align')).(0) in
      assert (align = back_align);
      let l, s, w = align.l, align.s, align.w in
      assert (l < A.length datas);
      assert (s < A.length datas.(l));
      assert (w < A.length datas.(l).(s));
   end) align_keys;
   
   let (a1, a2) = A.partition (fun align -> align.l = 0) (aoe align_keys) in
   assert (A.length a1 = A.length a2)   
;;
let filter_align d filter = 
   let num_aligns  = ref 0 in
   let num_aligns' = ref 0 in
   let keep l s w align = 
      let l' = align.l and s' = align.s and w' = align.w in
      let w_id  = d.sents.(l).(s).(w).id in
      let w_id' = d.sents.(l').(s').(w').id in
      filter (A.length d.lexs.(l).(w_id) <= 8) (A.length d.lexs.(l').(w_id') <= 8)
   in
   for l = 0 to A.length d.langs - 1 do
      for s = 0 to A.length d.sents.(l) - 1 do
         for w = 0 to A.length d.sents.(l).(s) - 1 do
            let align = d.sents.(l).(s).(w).align in
            let align' = A.filter (keep l s w) align in
            num_aligns  += A.length align;
            num_aligns' += A.length align';
            d.sents.(l).(s).(w).align <- align'
         done
      done
   done;
   brintf "NUM ALL ALIGNS = %d      NUM FILTERED ALIGNS = %d" !num_aligns !num_aligns';
;;   
let reduce_lex d n l : unit =
   brintf "REDUCING LEX...";
   let counts = Ht.create 15000 in
   Ht.clear counts;
   let pun_tag1 = A.find_item "PUN" d.maps.(l).pos in
   let pun_tag2 = A.find_item "." d.maps.(l).pos in
   let rest_tags = L.range 1 (A.length d.maps.(l).pos) |> (flip L.remove) pun_tag1 |> (flip L.remove) pun_tag2 |> aol in 
   for s = 0 to A.length d.sents.(l) - 1 do
      for w = 0 to A.length d.sents.(l).(s) - 1 do
         if not (d.lexs.(l).(d.sents.(l).(s).(w).id) = [|pun_tag1|]) && not (d.lexs.(l).(d.sents.(l).(s).(w).id) = [|pun_tag2|]) then
            Ht.incr counts d.sents.(l).(s).(w).id
      done
   done;
   let top = counts |> Ht.to_rev_sorted_array |> A.slice ~last:n |> A.map fst in
   if d.langs.(l) = "en" then
      brintf "reducing en lexicon to: %s" (A.to_str1 id (A.map (fun i -> utf16_to_str d.maps.(l).words.(i)) top));
   let top_hs = Hs.of_array top in
   for w = 0 to A.length d.lexs.(l) - 1 do
      if not (Hs.mem top_hs w) && not (d.lexs.(l).(w) = [|pun_tag1|]) && not (d.lexs.(l).(w) = [|pun_tag2|]) then
         d.lexs.(l).(w) <- rest_tags
   done;
   (* re-assign random tags *)
   for s = 0 to A.length d.sents.(l) - 1 do
      for w = 0 to A.length d.sents.(l).(s) - 1 do
         let word = d.sents.(l).(s).(w) in
         let psbl = d.lexs.(l).(word.id) in
         word.probs <- A.make (A.length psbl) 0.0;
         word.p_pos <- psbl.(Random.int (A.length psbl))
      done
   done
;;
let reduce_lex_freq d f l : unit =
   brintf "REDUCING LEX (remove words with frequency %d or less)..." f;
   let counts = Ht.create 15000 in
   Ht.clear counts;
   let tags = L.range 1 (A.length d.maps.(l).pos) |> aol in 
   for s = 0 to A.length d.sents.(l) - 1 do
      for w = 0 to A.length d.sents.(l).(s) - 1 do
         Ht.incr counts d.sents.(l).(s).(w).id
      done
   done;
   for w = 0 to A.length d.lexs.(l) - 1 do
      if (Ht.find counts w) <= f then
         d.lexs.(l).(w) <- tags;
   done;
   (* re-assign random tags *)
   for s = 0 to A.length d.sents.(l) - 1 do
      for w = 0 to A.length d.sents.(l).(s) - 1 do
         let word = d.sents.(l).(s).(w) in
         let psbl = d.lexs.(l).(word.id) in
         let probs = A.make (A.length psbl) 0.0 in
         word.probs <- probs;
         word.p_pos <- psbl.(Random.int (A.length psbl))
      done
   done
;;

let config           = Ht.of_file id id Sys.argv.(1)

let read_data : data = 
   let langs         = splita (Ht.find config "languages") in
   let lex_files     = splita (Ht.find config "lexica") in
   let data_files    = splita (Ht.find config "data") in
   let align_file    = (splita (Ht.find config "align")).(0) in
   let testpoints    = A.map ios (splita (Ht.find config "testpoints")) in
   let endpoints     = A.mapi (fun i s -> if s = "None" then testpoints.(i) else ios s) (splita (Ht.find config "endpoints")) in
   let suppoints     = A.map (fun s -> if s = "None" then None else Some (Pair.map ios (S.split s "--"))) (splita (Ht.find config "suppoints")) in
   let anneal        = bos (splita (Ht.find config "anneal")).(0) in
   let filter        = match (splita (Ht.find config "filter")).(0) with 
                        "or" -> Some (||) | "and" -> Some (&&) | "none" -> None | _ -> failwith ("bad filter in config") in
   let lengths       = L.map A.length [langs; lex_files; data_files] in
   assert (L.all_same lengths);
   assert (L.hd lengths = A.length testpoints);
   assert (A.length langs = 2);
   let n             = L.hd lengths in
   brintf "Reading lexicon files...";
   let lexs          = A.map read_lex_file lex_files in
   brintf "Reading data files...";
   let datas         = A.map read_data_file data_files in
   let align         = read_align_file align_file langs in
   brintf "Checking alignments...";
   check_align align datas langs;
   remove_crossing_alignments align;
   check_align align datas langs;
   brintf "Compiling data...";
   let compiled_data = A.init n (fun i -> compile_data i lexs.(i) datas.(i) align) in
   let res = {
      langs = langs;
      sents = A.init n (fun i -> match compiled_data.(i) with x,_,_ -> x);
      maps  = A.init n (fun i -> match compiled_data.(i) with _,x,_ -> x);
      lexs  = A.init n (fun i -> match compiled_data.(i) with _,_,x -> x);
      n     = foi ( Ht.length align / 2 );
      testpoints  = testpoints;
      endpoints   = endpoints;
      suppoints   = suppoints;
      anneal      = anneal
   } in
   let reduce1, reduce2 = Pair.map bos (S.split (Ht.find config "reduce_lex") " ") in
   let freq1, freq2 = S.split (Ht.find config "frequency") " " in
   if reduce1 then if freq1 = "none" then reduce_lex res 100 0 else reduce_lex_freq res (ios freq1) 0;
   if reduce2 then if freq2 = "none" then reduce_lex res 100 1 else reduce_lex_freq res (ios freq2) 1;
   if (reduce1 || reduce2) && filter != None then filter_align res (Opt.get filter);
   res
;;
let init_model (d : data) : model =
   let num_langs  = A.length d.langs in
   let m = { 
      theta        = A.make num_langs (fos (splita (Ht.find config "theta_o")).(0));
      eta          = A.make num_langs (fos (splita (Ht.find config "eta_o")).(0));
      zeta         = (fos (splita (Ht.find config "zeta_o")).(0));
      cpcount      = A.init_2D (A.length d.maps.(0).pos) (A.length d.maps.(1).pos) (fun _ _ -> 0.);
      emcount      = A.init num_langs (fun l -> A.make_matrix (A.length d.maps.(l).pos) (A.length d.maps.(l).words) 0.);
      trcount      = A.init num_langs (fun l -> let nt = A.length d.maps.(l).pos in A.make_3D nt nt nt 0.);

      bicount      = A.init num_langs (fun l -> let nt = A.length d.maps.(l).pos in A.make_matrix nt nt 0.);
      monocount    = A.init num_langs (fun l -> A.make (A.length d.maps.(l).pos) 0.);
      tagwordcount = A.init num_langs (fun l -> A.make (A.length d.maps.(l).pos) 0.);

      bicountx      = A.init num_langs (fun l -> let nt = A.length d.maps.(l).pos in A.make_matrix nt nt 0.);
      monocountx    = A.init num_langs (fun l -> A.make (A.length d.maps.(l).pos) 0.);
      emcountx      = A.init num_langs (fun l -> A.make_matrix (A.length d.maps.(l).pos) (A.length d.maps.(l).words) 0.);
      trcountx      = A.init num_langs (fun l -> let nt = A.length d.maps.(l).pos in A.make_3D nt nt nt 0.);
   } in
   m
;;

let d                = read_data
let m                = init_model d
let anneal_stop      = 1000
let get_schedule     = get_get_schedule 0.5 1.25
let schedule         = get_schedule 0 anneal_stop
(* let out_dir          = (splita (Ht.find config "test_out")).(0);;
assert (out_dir = "test_out");; *)

let prev_two_tags l s w =
   let sent = d.sents.(l).(s) in
   let len = A.length sent in
   try match w with
   | 0                  -> (0, 0)
   | 1                  -> (0, sent.(w-1).p_pos)
   | x when x = len + 1 -> (sent.(w-2).p_pos, 0) 
   | _                  -> (sent.(w-2).p_pos, sent.(w-1).p_pos)
   with Invalid_argument "index out of bounds" -> 
      failwith(Format.sprintf "prev_two_tags:  len = %d, w = %d" len w) 
;;
let next_two_tags l s w = 
   let sent = d.sents.(l).(s) in
   let len = A.length sent in
   try match w with
   | x when x = len - 1    -> (0, 0)
   | x when x = len - 2    -> (sent.(w+1).p_pos, 0)
   | _                     -> (sent.(w+1).p_pos, sent.(w+2).p_pos)
   with Invalid_argument "index out of bounds" -> 
      failwith(Format.sprintf "next_two_tags:  len = %d, w = %d" len w) 
;;
let init_sup () : unit =
   A.iteri (begin fun l opt -> match opt with
      | Some (a, b) ->
         for s = a to min b (A.length d.sents.(l) - 1) do
            let sent = d.sents.(l).(s) in
            for w = 0 to A.length sent - 1 do
               sent.(w).p_pos <- sent.(w).g_pos;
               sent.(w).x_pos <- sent.(w).g_pos
            done
         done
      | None -> ()
   end) d.suppoints
;;
let init_counts () : unit =
   (* initially set all counts to 0 *)
   for l = 0 to A.length d.langs - 1 do
      for t = 0 to A.length m.trcount.(l) -1 do
         A.fill_all_matrix m.trcount.(l).(t) 0.0
      done;
      A.fill_all_matrix m.emcount.(l) 0.0 ;
      A.fill_all_matrix m.bicount.(l) 0.0 ;
      A.fill_all m.monocount.(l) 0.0
   done;
   (* updates counts while iterating through data *)
   let alns = ref 0.0 in
   for l = 0 to A.length d.langs - 1 do
      for s = 0 to d.endpoints.(l) do
         let sen = d.sents.(l).(s) in
         let slen = A.length sen - 1 in
         A.incr' m.bicount.(l).(0) 0 ;
         for w = 0 to slen do
            let word = sen.(w) in
            let pp_tag, p_tag = prev_two_tags l s w in
            let tag = word.p_pos in
            A.incr' m.trcount.(l).(pp_tag).(p_tag) tag ;
            A.incr' m.emcount.(l).(tag) word.id ;
            A.incr' m.monocount.(l) tag ;
            A.incr' m.bicount.(l).(p_tag) tag ;
            if w == slen then begin
               A.incr' m.trcount.(l).(p_tag).(tag) 0 ;
               A.incr' m.trcount.(l).(tag).(0) 0 ;
               A.incr' m.bicount.(l).(tag) 0 ;
            end;
            assert (A.length word.align = 0 || A.length word.align = 1);
            for a = 0 to A.length word.align - 1 do
               if l = 0 then begin
                  let l'   = word.align.(a).l in
                  assert (l' = 1);
                  let s'   = word.align.(a).s in
                  let w'   = word.align.(a).w in
                  let tag' = d.sents.(l').(s').(w').p_pos in
                  let back_align = d.sents.(l').(s').(w').align in
                  if (A.length back_align = 0) then
                     brintf "no back align!!! s = %d   w = %d   s' = %d   w' = %d" s w s' w';
                  assert (A.length back_align = 1);
                  assert (back_align.(0).l = 0 && back_align.(0).w = w && back_align.(0).s = s);
                  A.incr' m.cpcount.(tag) tag';
                  alns := !alns +. 1.0;
               end
            done
         done
      done;
      for w = 0 to A.length d.lexs.(l) - 1 do
         for t = 0 to A.length d.lexs.(l).(w) -1 do
            A.incr' m.tagwordcount.(l) d.lexs.(l).(w).(t)
         done;
      done;
   done;
   
   (*d.n <- !alns;*)
   brintf "d.n=%f alns=%f" d.n !alns;
;;

init_sup ();;
init_counts ();;

(*************************************************************************************)
let is_sup l s = 
   match d.suppoints.(l) with
      | None ->      false
      | Some (a, b) ->  s >= a && s <= b
;;
let prev_two_xtags l s w =
   let sent = d.sents.(l).(s) in
   let len = A.length sent in
   try match w with
      | 0                  -> (0, 0)
      | 1                  -> (0, sent.(w-1).x_pos)
      | x when x = len + 1 -> (sent.(w-2).x_pos, 0)
      | _                  -> (sent.(w-2).x_pos,
   sent.(w-1).x_pos)
   with Invalid_argument "index out of bounds" ->
      failwith(Format.sprintf "prev_two_max_tags:  len = %d, w = %d" len w)
;;
let next_two_xtags l s w =
   let sent = d.sents.(l).(s) in
   let len = A.length sent in
   try match w with
      | x when x = len - 1    -> (0, 0)
      | x when x = len - 2    -> (sent.(w+1).x_pos, 0)
      | _                     -> (sent.(w+1).x_pos,
   sent.(w+2).x_pos)
   with Invalid_argument "index out of bounds" ->
      failwith(Format.sprintf "next_two_max_tags:  len = %d, w = %d" len w)
;;
let choose_xtags () =
   let num_langs = A.length d.langs in
   for l = 0 to num_langs - 1 do
      for s = 0 to d.endpoints.(l) do
         let sen = d.sents.(l).(s) in
         for w = 0 to A.length sen - 1 do
            let word  = sen.(w) in
            let probs = word.probs in
            let maxi  = ref 0 in
            for i = 1 to A.length probs - 1 do
               if probs.(i) > probs.(!maxi) then
                  maxi := i;
            done;
            word.x_pos <- d.lexs.(l).(word.id).(!maxi);
         done;
      done;
   done;
;;
let update_xcounts () : unit =
   (* initially set all counts to 0 *)
   for l = 0 to A.length d.langs - 1 do
      A.fill_all_matrix m.bicountx.(l) 0.0 ;
      A.fill_all_matrix m.emcountx.(l) 0.0 ;
      for t = 0 to A.length d.maps.(l).pos - 1 do
         A.fill_all_matrix m.trcountx.(l).(t) 0.0 ;
      done;
      A.fill_all m.monocountx.(l) 0.0;
   done;
   (* updates counts while iterating through data *)
   for l = 0 to A.length d.langs - 1 do
      for s = 0 to d.endpoints.(l) do
         let sen = d.sents.(l).(s) in
         let slen = A.length sen - 1 in
         A.incr' m.bicountx.(l).(0) 0 ;
         for w = 0 to slen do
            let word = sen.(w) in
            let pp_tag, p_tag = prev_two_xtags l s w in
            let tag = word.x_pos in
            A.incr' m.trcountx.(l).(pp_tag).(p_tag) tag ;
            A.incr' m.emcountx.(l).(tag) word.id ;
            A.incr' m.monocountx.(l) tag ;
            if w == slen then begin
               A.incr' m.trcountx.(l).(p_tag).(tag) 0 ;
               A.incr' m.trcountx.(l).(tag).(0) 0 ;
            end;    
            A.incr' m.bicountx.(l).(p_tag) tag ;
         done;
      done;
   done;
;;
let viterbi l s x : unit =
 (* sentence level viterbi  *)
  let sen = d.sents.(l).(s) in
  let size = A.length sen in
  let tagset_size = A.length d.maps.(l).pos in
  let pmat = A.make_matrix tagset_size tagset_size neg_infinity in
  let bmat = A.make_matrix tagset_size tagset_size 0 in
  
   let (trcount,bicount,emcount,monocount) =
      if x then (m.trcountx, m.bicountx, m.emcountx, m.monocountx)
      else (m.trcount, m.bicount, m.emcount, m.monocount) in

  let prob_matrix_arr = A.init (size + 1) (fun _ -> A.copy_matrix pmat) in
  let back_matrix_arr = A.init (size + 1) (fun _ -> A.copy_matrix bmat) in

   prob_matrix_arr.(0).(0).(0) <- 0.0;
   back_matrix_arr.(0).(0).(0) <-  -1;
  
   let alpha   = m.eta.(l) in
   let beta    = m.theta.(l) in
   let t_alpha = alpha *. float tagset_size in
   for w = 1 to size do
      let word = sen.(w-1).id in
      let wtags = d.lexs.(l).(word) in
      for i = 0 to A.length wtags - 1 do
         let tag = wtags.(i) in
         for tag' = 0 to tagset_size - 1 do
            let maxp = ref neg_infinity in
            for tag'' = 0 to tagset_size - 1 do
               let p = prob_matrix_arr.(w-1).(tag'').(tag') +.
                  log10 ( ( trcount.(l).(tag'').(tag').(tag) +. alpha ) /. ( bicount.(l).(tag'').(tag') +. t_alpha ) ) +.
                  log10 ( ( emcount.(l).(tag).(word) +. beta ) /. ( monocount.(l).(tag) +. beta *.  m.tagwordcount.(l).(tag) ) ) in
               if p > !maxp then begin
                  maxp := p ;
                  prob_matrix_arr.(w).(tag').(tag) <- p ;
                  back_matrix_arr.(w).(tag').(tag) <- tag''
               end
            done
         done
      done
   done;

  let thistag = ref 0 in
  let p_tag = ref 0 in
  let max = ref (log10 0.0) in
  for tag' = 0 to tagset_size - 1 do
    for tag = 0 to tagset_size - 1 do
       let p = prob_matrix_arr.(size).(tag').(tag) +. 
            log10 ( ( trcount.(l).(tag').(tag).(0) +.  alpha) /. ( bicount.(l).(tag').(tag) +. t_alpha ) )in
       if p > !max then begin
          thistag := tag ;
          p_tag := tag';
          max := p ;
       end
    done;
   done;

   for w = size downto 1 do

      if x then 
         sen.(w-1).x_pos <- !thistag
      else
         sen.(w-1).p_pos <- !thistag;

     let tempthistag = !thistag in
     thistag := !p_tag;
     if w > 1 then   
        p_tag := back_matrix_arr.(w).(!p_tag).(tempthistag)
   done
;;
let annotate_testdata x : unit =
   let langcount = A.length d.langs in
   if x then
      update_xcounts ();
   for l = 0 to langcount - 1 do
      for s = d.testpoints.(l) + 1 to A.length d.sents.(l) - 1 do
         viterbi l s x
      done;
   done;  
;;

let emit_prob l tag word = 
   let count      = m.emcount.(l).(tag).(word) in
   let tag_count  = m.monocount.(l).(tag) in
   assert (count >= 0.);
   assert (tag_count >= 0.);
   (count +. m.theta.(l)) /. (tag_count +. m.tagwordcount.(l).(tag) *. m.theta.(l))
;;
let trans_prob l pp_tag p_tag tag = 
   let prev_count = m.bicount.(l).(pp_tag).(p_tag) in
   let count      = m.trcount.(l).(pp_tag).(p_tag).(tag) in
   assert (count >= 0.);
   assert (prev_count >= 0.);
   (count +. m.eta.(l)) /. (prev_count +. m.eta.(l) *. foi (A.length d.maps.(l).pos)) 
;;
let cpl_prob l tag l' tag' = 
   let t, t' = match (l,l') with
      | (0,1)  -> tag, tag'
      | (1,0)  -> tag', tag 
      | _ -> failwith "bad bad"
   in 
   let count = m.cpcount.(t).(t') in
   if count < 0. then
      brintf "cpcount.(%d=%s).(%d=%s) = %d" t d.maps.(0).pos.(t) t' d.maps.(1).pos.(t') (iof count);
   assert (count >= 0.);
   assert (d.n > 0.);
   (count +. m.zeta) /. (d.n +. m.zeta *. foi (A.length d.maps.(0).pos) *. foi (A.length d.maps.(1).pos))  
;;

let compute_tag_prob l s w = 
   let sent = d.sents.(l).(s) in
   let tag = 
      if w > A.length sent - 1 then 0 else sent.(w).p_pos in
   let pp_tag, p_tag = prev_two_tags l s w in
   trans_prob l pp_tag p_tag tag
;;
let compute_tagpair_prob_aux l pp_tag p_tag tag l' pp_tag' p_tag' tag' = 
   let tsize   = A.length d.maps.(l).pos in
   let tsize'  = A.length d.maps.(l').pos in
   let p    = ref 0.0 in
   let norm = ref 0.0 in
   for t = 0 to tsize - 1 do
      for t' = 0 to tsize' - 1 do
         let score = trans_prob l pp_tag p_tag t *.
                     trans_prob l' pp_tag' p_tag' t' *.
                     cpl_prob l t l' t'
         in
         norm := !norm +. score;
         if t = tag && t' = tag' then
            p := score
      done
   done;
   !p /. !norm
;;


(* computes the generative probability of the current setting of the tag or tag-pair if it is aligned *)
let compute_tagpair_prob l s w pp_tag p_tag tag =
   let sent = d.sents.(l).(s) in
   if tag = 0 || A.length sent.(w).align = 0 then
      trans_prob l pp_tag p_tag tag
   else begin
      let align = sent.(w).align.(0) in
      let l' = align.l in
      let s' = align.s in
      let w' = align.w in
      let tag' = d.sents.(l').(s').(w').p_pos in
      let pp_tag', p_tag'  = prev_two_tags l' s' w' in
      compute_tagpair_prob_aux l pp_tag p_tag tag l' pp_tag' p_tag' tag'
   end
;;
(* gets the (unnormalized) sampling probabilities of all possible tag pairs assuming neither tag is observed *)
let get_tagpair_sampling_probs1 word word' sent sent' l s w l' s' w' pp_tag p_tag n_tag nn_tag pp_tag' p_tag' n_tag' nn_tag' r =
   let tags    = d.lexs.(l).(word.id) in
   let tags'   = d.lexs.(l').(word'.id) in
   let tsize   = A.length tags in
   let tsize'  = A.length tags' in
   let sent_len   = A.length sent in
   let n_align    = if w > sent_len - 2 then [||] else sent.(w + 1).align in
   let nn_align   = if w > sent_len - 3 then [||] else sent.(w + 2).align in
   let res     = A.make (tsize * tsize') 0.0 in
   let e = if d.anneal then schedule.(r) else 1.0 in

   for idx = 0 to tsize - 1 do
      for idx' = 0 to tsize' - 1 do
         let t    = tags.(idx) in
         let t'   = tags'.(idx') in
         word.p_pos     <- t;
         word'.p_pos    <- t';
         let n_prb'  = compute_tagpair_prob l' s' (w' + 1) p_tag' t' n_tag' in
         let nn_prb' = compute_tagpair_prob l' s' (w' + 2) t' n_tag' nn_tag' in
         let n_prb   = 
            if A.length n_align > 0 && n_align.(0).s = s' && (n_align.(0).w = (w' + 1) || n_align.(0).w = (w' + 2)) then 1.0 
            else compute_tagpair_prob l s (w + 1) p_tag t n_tag 
         in
         let nn_prb  = 
            if A.length nn_align > 0 && nn_align.(0).s = s' && (nn_align.(0).w = (w' + 1) || nn_align.(0).w = (w' + 2)) then 1.0
            else compute_tagpair_prob l s (w + 2) t n_tag nn_tag
         in
         let prob =  trans_prob l pp_tag p_tag t *. 
                     trans_prob l' pp_tag' p_tag' t' *.
                     n_prb *. 
                     nn_prb *.
                     n_prb' *.
                     nn_prb' *.
                     emit_prob l t word.id *.
                     emit_prob l' t' word'.id *.
                     cpl_prob l t l' t'
         in
         res.(idx * tsize' + idx')  <- prob ** e;
         word.probs.(idx)           <- word.probs.(idx) +. prob;
         word'.probs.(idx')         <- word'.probs.(idx') +. prob
      done
   done;
   res
;; 
(* gets the (unnormalized) sampling probabilities of all possible first-language tags assuming the second-language tag is observed *)
let get_tagpair_sampling_probs2 word word' l s w l' s' w' pp_tag p_tag n_tag nn_tag r =
   (* let pp_tag, p_tag = prev_two_tags l s w in *)
   let tags    = d.lexs.(l).(word.id) in
   let tsize   = A.length tags in
   let res     = A.make tsize 0.0 in
   let e       = if d.anneal then schedule.(r) else 1.0 in
   let t'      = word'.p_pos in

   for idx = 0 to tsize - 1 do
      let t = tags.(idx) in
      let prob =  trans_prob l pp_tag p_tag t *. 
                  compute_tagpair_prob l s (w + 1) p_tag t n_tag *.
                  compute_tagpair_prob l s (w + 2) t n_tag nn_tag *.
                  emit_prob l t word.id *.
                  cpl_prob l t l' t' 
      in         
      res.(idx)         <- prob ** e;
      word.probs.(idx)  <- word.probs.(idx) +. prob
   done;
   res
;; 
let incr_counts l word pp_tag p_tag n_tag nn_tag : unit = 
   let tag = word.p_pos in
   A.incr' m.trcount.(l).(pp_tag).(p_tag) tag ;
   A.incr' m.trcount.(l).(p_tag).(tag) n_tag ;
   A.incr' m.trcount.(l).(tag).(n_tag) nn_tag ;
   A.incr' m.bicount.(l).(p_tag) tag ;
   A.incr' m.bicount.(l).(tag) n_tag ;
   A.incr' m.emcount.(l).(tag) word.id ;
   A.incr' m.monocount.(l) tag ;
;;
let decr_counts l word pp_tag p_tag n_tag nn_tag : unit =
   let tag = word.p_pos in
   A.decr' m.trcount.(l).(pp_tag).(p_tag) tag ;
   A.decr' m.trcount.(l).(p_tag).(tag) n_tag ;
   A.decr' m.trcount.(l).(tag).(n_tag) nn_tag ;
   A.decr' m.bicount.(l).(p_tag) tag ;
   A.decr' m.bicount.(l).(tag) n_tag ;
   A.decr' m.emcount.(l).(tag) word.id ;
   A.decr' m.monocount.(l) tag ;
;;
let sample_tag word l s w pp_tag p_tag n_tag nn_tag r : unit = 
   
   assert (not (is_sup l s));
   let lex  = d.lexs.(l).(word.id) in
   let len = A.length lex in
   let p = A.make len 0.0 in
   
   (* Decrement affected counts *)
   decr_counts l word pp_tag p_tag n_tag nn_tag;
   
   (* aggregate probabilities *)
   let e = if d.anneal then schedule.(r) else 1.0 in
   for i = 0 to len - 1 do
      let tag = lex.(i) in
      word.p_pos <- tag;
      p.(i) <- trans_prob l pp_tag p_tag tag 
            *. compute_tagpair_prob l s (w+1) p_tag tag n_tag 
            *. compute_tagpair_prob l s (w+2) tag n_tag nn_tag
            *. emit_prob l tag word.id ;
      word.probs.(i) <- word.probs.(i) +. p.(i);
      p.(i) <- p.(i) ** e
   done;

   (* Sample and set the tag *)
   let i = Gsl_randist.discrete rng (Gsl_randist.discrete_preproc p) in
   word.p_pos <- lex.(i);

   (* Increment the counts for new tag *)
   incr_counts l word pp_tag p_tag n_tag nn_tag;
;;
let sample_tag_pair sent word align l s w pp_tag p_tag n_tag nn_tag r : unit =
   let l' = align.l in
   let s' = align.s in
   let w' = align.w in
   let sent' = d.sents.(l').(s') in
   let word' = sent'.(w') in
   let pp_tag', p_tag'  = prev_two_tags l' s' w' in
   let n_tag', nn_tag'  = next_two_tags l' s' w' in
   let thistag  = word.p_pos in
   let thistag' = word'.p_pos in
   let sup  = is_sup l s in
   let sup' = is_sup l' s' in 
   assert (not sup);
   
   (* Decrement affected counts *)
   if not sup then
      decr_counts l word pp_tag p_tag n_tag nn_tag;
   if not sup' then
      decr_counts l' word' pp_tag' p_tag' n_tag' nn_tag';
   if l = 0 then begin
      let count = m.cpcount.(thistag).(thistag') in
      assert (count > 0.);
      A.decr' m.cpcount.(thistag) thistag'
   end 
   else begin
      assert (l = 1);
      let count = m.cpcount.(thistag').(thistag) in
      assert (count > 0.);
      A.decr' m.cpcount.(thistag') thistag
   end;
      
   let tags   = d.lexs.(l).(word.id) in
   let tags'  = d.lexs.(l').(word'.id) in
   let tsize' = A.length tags' in

   (* Sample and set the tag pair *)
   let p = match sup, sup' with
      | false, false -> get_tagpair_sampling_probs1 word word' sent sent' l s w l' s' w' pp_tag p_tag n_tag nn_tag pp_tag' p_tag' n_tag' nn_tag' r 
      | false, true  -> get_tagpair_sampling_probs2 word word' l s w l' s' w' pp_tag p_tag n_tag nn_tag r 
      | true, false  -> failwith "trying to samply supervised tag!"
      | _ -> failwith "trying to sample tag-pair when both are supervised!"
   in
   let i = Gsl_randist.discrete rng (Gsl_randist.discrete_preproc p) in 
   begin match sup, sup' with
      | false, false ->
         word.p_pos  <- tags.(i/tsize');
         word'.p_pos <- tags'.(i mod tsize')
      | false, true -> 
         word.p_pos  <- tags.(i)
      | true, false -> failwith "trying to sample supervised tag!"
      | _ -> failwith "trying to sample tag-pair when both are supervised!" 
   end;
   let newtag  = word.p_pos in
   let newtag' = word'.p_pos in

   (* Increment affected counts *)
   if not sup then
      incr_counts l word pp_tag p_tag n_tag nn_tag;
   if not sup' then
      incr_counts l' word' pp_tag' p_tag' n_tag' nn_tag';
   if l = 0 then
      A.incr' m.cpcount.(newtag) newtag'
   else begin
      assert (l = 1);
      A.incr' m.cpcount.(newtag') newtag
   end
;;

let asc_fact par count =
   let fact = ref 0.0 in
   for i = 0 to iof (count) - 1 do
      fact := !fact +. log(par +. foi i)
   done;
   !fact
;;
let compute_ll () = 
   let num     = ref 0.0 in
   let denom   = ref 0.0 in
   let tsize1 = A.length d.maps.(0).pos in
   let tsize2 = A.length d.maps.(1).pos in
   (* deal with couplings *)
   denom := !denom +. asc_fact (foi tsize1 *. foi tsize2 *. m.zeta) d.n;
   for t = 0 to tsize1 - 1 do
      for t' = 0 to tsize2 - 1 do
         num := !num +. asc_fact m.zeta m.cpcount.(t).(t')
      done
   done;
         
   for l = 0 to A.length d.langs - 1 do
      let theta   = m.theta.(l) in
      let eta     = m.eta.(l) in
      let tsize   = A.length d.maps.(l).pos in
      let emcount    = m.emcount.(l) in
      let monocount  = m.monocount.(l) in
      let tagwordcount  = m.tagwordcount.(l) in
      let bicount       = m.bicount.(l) in
      let trcount       = m.trcount.(l) in
      
      for t = 0 to tsize - 1 do
         denom := !denom +. asc_fact (tagwordcount.(t) *. theta) monocount.(t);
         A.iter (fun count -> num := !num +. asc_fact theta count) emcount.(t);
         for t' = 0 to tsize - 1 do
            if (t = 0 || t' > 0) then begin
               denom := !denom +. asc_fact (foi tsize *. eta) bicount.(t).(t');
               for t'' = 0 to tsize - 1 do
                  num := !num +. asc_fact eta trcount.(t).(t').(t'')
               done
            end
         done
      done
   done;
   !num -. !denom
;;       
let accept p p' q q' : bool = 
   (*assuming both likelihood and prob are in natural log *)
   (* p = old prob; p' = new prob; q = old proposal; q' = new proposal *)
  Gsl_rng.uniform rng <= exp (( p' +. q ) -. ( p +. q'))
;;
let estimate_theta l : unit =
   let oldtheta = m.theta.(l) in
   let oldmean  = oldtheta in
   let oldvar   = oldmean *. 0.1 in
   let newtheta = oldmean +. Gsl_randist.gaussian rng oldvar in
   let newmean  = newtheta in
   let newvar   = newmean *. 0.1 in
   let q        = log(Gsl_randist.gaussian_pdf (oldtheta -. newmean) newvar) in
   let q'       = log(Gsl_randist.gaussian_pdf (newtheta -. oldmean) oldvar) in
   let p      = ref 0.0 in
   let p'     = ref 0.0 in
   let numo   = ref 0.0 in
   let numn   = ref 0.0 in
   let denomo = ref 0.0 in
   let denomn = ref 0.0 in
   for t = 0 to A.length d.maps.(l).pos - 1 do
      A.iter (fun count -> numo := !numo +. asc_fact oldtheta count ) m.emcount.(l).(t);
      A.iter (fun count -> numn := !numn +. asc_fact newtheta count ) m.emcount.(l).(t);
      denomo := asc_fact (m.tagwordcount.(l).(t) *. oldtheta) m.monocount.(l).(t);
      denomn := asc_fact (m.tagwordcount.(l).(t) *. newtheta) m.monocount.(l).(t);
      p      := !p +. ( !numo -. !denomo );
      p'     := !p' +. ( !numn -. !denomn );
      numo := 0.; numn := 0.; denomo := 0.; denomn := 0.;
   done; 
   if accept !p !p' q q' then
      m.theta.(l) <- newtheta;
;;
let estimate_eta l = 
   let oldeta  = m.eta.(l) in
   let oldmean = oldeta in
   let oldvar  = oldmean *. 0.1 in
   let neweta  = oldmean +. Gsl_randist.gaussian rng oldvar in
   let newmean = neweta in
   let newvar  = newmean *. 0.1 in
   let q       = log(Gsl_randist.gaussian_pdf (oldeta -. newmean) newvar) in
   let q'      = log(Gsl_randist.gaussian_pdf (neweta -. oldmean) oldvar) in
   let numo   = ref 0.0 in
   let numn   = ref 0.0 in
   let denomo = ref 0.0 in
   let denomn = ref 0.0 in
   let tsize  = A.length d.maps.(l).pos in
   for t = 0 to tsize - 1 do
      for t' = 0 to tsize - 1 do
         if (t = 0 || t' > 0) then begin
            denomo := !denomo +. asc_fact (foi tsize *. oldeta) m.bicount.(l).(t).(t');
            denomn := !denomn +. asc_fact (foi tsize *. neweta) m.bicount.(l).(t).(t');
            for t'' = 0 to tsize - 1 do
               numo := !numo +. asc_fact oldeta m.trcount.(l).(t).(t').(t'');
               numn := !numn +. asc_fact neweta m.trcount.(l).(t).(t').(t'')
            done
         end
      done
   done;
   let p  = !numo -. !denomo in
   let p' = !numn -. !denomn in 
   if accept p p' q q' then
      m.eta.(l) <- neweta;
;;
let estimate_zeta () = 
   let oldzeta  = m.zeta in
   let oldmean  = oldzeta in
   let oldvar   = oldmean *. 0.1 in
   let newzeta  = oldmean +. Gsl_randist.gaussian rng oldvar in
   let newmean  = newzeta in
   let newvar   = newmean *. 0.1 in
   let q        = log(Gsl_randist.gaussian_pdf (oldzeta -. newmean) newvar) in
   let q'       = log(Gsl_randist.gaussian_pdf (newzeta -. oldmean) oldvar) in
   let tsize1   = A.length d.maps.(0).pos in
   let tsize2   = A.length d.maps.(1).pos in
   let numo   = ref 0.0 in
   let numn   = ref 0.0 in
   let denomo = asc_fact (foi tsize1 *. foi tsize2 *. oldzeta) d.n in
   let denomn = asc_fact (foi tsize1 *. foi tsize2 *. newzeta) d.n in
   for t = 0 to tsize1 - 1 do
      for t' = 0 to tsize2 - 1 do
         numo := !numo +. asc_fact oldzeta m.cpcount.(t).(t');
         numn := !numn +. asc_fact newzeta m.cpcount.(t).(t')
      done
   done;
   let p  = !numo -. denomo in
   let p' = !numn -. denomn in 
   if accept p p' q q' then
      m.zeta <- newzeta;
;;
let test ?fn (final : bool) (x : bool) =
   (* returns an array of accuracy values, one for each language *)
   if x then
      choose_xtags ();
   if final then
      annotate_testdata x;
   let save = fn != None in
   let langcount = A.length d.langs in
   let res = A.make langcount 0.0 in
   for l = 0 to langcount - 1 do
      let out_str = Buffer.create (if save then 10 else 50000) in
      let good  = ref 0.0 in
      let total = ref 0.0 in
      let pun   = A.find_item "PUN" d.maps.(l).pos in
      let sm    = A.find_item "." d.maps.(l).pos in
      let start, stop = match final with
         | true   -> d.testpoints.(l) + 1, A.length d.sents.(l) -1
         | false  -> 0, d.endpoints.(l) in
      for s = start to stop do
         for w = 0 to A.length d.sents.(l).(s) - 1 do
            let p_pos = match x with
               | true   -> d.sents.(l).(s).(w).x_pos
               | false  -> d.sents.(l).(s).(w).p_pos in
            if not ( p_pos = pun || p_pos = sm ) then begin
               total := !total +. 1. ;
               if ( p_pos == d.sents.(l).(s).(w).g_pos ) then
                  good := !good +. 1. ;
            end;
            if save then 
               Buffer.add_string out_str (d.maps.(l).pos.(p_pos) ^ " ");
         done;
         if save then
            Buffer.add_string out_str "\n";
      done;
      res.(l) <- !good /. !total *. 100.0;
      if save then
         output_file ~filename:(Opt.get fn).(l) ~text:(Buffer.contents out_str);
   done;
   res
;;

let sample r : unit =
   brintf "SAMPLING ROUND = %d" r;
   for l = 0 to A.length d.langs - 1 do
      let sents = d.sents.(l) in
      for s = 0 to d.endpoints.(l) do
         if not (is_sup l s) then begin
            let sent = sents.(s) in
            let len_sent = A.length sent in
            let pp_tag  = ref 0 in
            let p_tag   = ref 0 in
            for w = 0 to len_sent - 1 do
               let word    = sent.(w) in
               let align   = word.align in
               let n_tag   = if w < len_sent - 1 then sent.(w+1).p_pos else 0 in
               let nn_tag  = if w < len_sent - 2 then sent.(w+2).p_pos else 0 in
               (if A.length align = 0 then
                  sample_tag word l s w !pp_tag !p_tag n_tag nn_tag r
               else
                  sample_tag_pair sent word align.(0) l s w !pp_tag !p_tag n_tag nn_tag r);
               pp_tag := !p_tag;
               p_tag  := word.p_pos
            done
         end
      done
   done;
   if r mod 10 = 0 then begin
      estimate_zeta (); 
      for l = 0 to A.length d.langs - 1 do
         estimate_theta l;
         estimate_eta l
       done
   end;
;;
let main () = 

   let stable     = 1000 in
   let final_stop = 200 in
   let eps        = 0.001 in
   let converged  = ref false in
   let r          = ref 0 in
   let nll_q      = Queue.create () in   

   let res = test false false in
   A.iteri (fun l r -> brintf "INITIAL accuracy for %s = %f" d.langs.(l) r) res;
   for r = 0 to 200 do 
      estimate_zeta ();
      for l = 0 to A.length d.langs - 1 do
         estimate_theta l;
         estimate_eta l
      done
   done;
   
   while ((not d.anneal && not !converged && !r < final_stop) || (d.anneal && !r < anneal_stop)) do
      sample !r;
      
      let nll = -1.0 *. compute_ll () in
      if !r mod 50 = 0 then begin
         let restrain   = test false false in
         let restrainx  = test false true in
         let restest    = test true false in
         let restestx   = test true true in
         A.iteri begin 
            fun l lang -> brintf "      %s     nll = %f  train(x) = %.2f %.2f test(x) = %.2f %.2f" lang nll restrain.(l) restrainx.(l) restest.(l) restestx.(l)
         end d.langs;
      end;

      Queue.push nll nll_q;
      if !r >= stable then begin
         let p_nll = Queue.pop nll_q in      
         let diff = abs_float (nll -. p_nll) in
         if diff < eps *. p_nll then 
            converged := true;
      end;

      if !r mod 50 = 0 then begin 
         (* print_endline "compacting GC"; *)
         (* let tm = Unix.gettimeofday () in *)
         Gc.compact ();
         (* let diff = Unix.gettimeofday () -. tm in
         brintf "done (%.4f)" diff; *)
      end; 
      incr r
   done;

   let fn1 = Sys.argv.(2) ^ "-" ^ d.langs.(0) ^ "_" ^ d.langs.(1) in
   let fn2 = Sys.argv.(2) ^ "-" ^ d.langs.(1) ^ "_" ^ d.langs.(0) in
   let nll        = -1.0 *. compute_ll () in
   let restrain   = test false false in
   let restrainx  = test false true in
   let restest    = test true false in
   let restestx   = test ~fn:[|fn1; fn2|] true true in
   A.iteri begin
      fun l lang -> brintf "      %s     nll = %f  train(x) = %.2f %.2f test(x) = %.2f %.2f" lang nll restrain.(l) restrainx.(l) restest.(l) restestx.(l)
   end d.langs
;;

main ()































